/* $Id: comp.c,v 1.3.1.5 2002/04/23 12:05:27 deton Exp $ */

/* comp.c from tserv-0.2 <http://www.tcp-ip.or.jp/~tagawa/archive/>.
 * modified for Win.
 */

/*
 *                    Copyright (C) Shingo NISHIOKA, 1992
 *                       nishioka@sanken.osaka-u.ac.jp
 *            Everyone is permitted to do anything on this program
 *         including copying, transplanting, debugging, and modifying.
 */

/*
  ̃vǑ^, tc.el ܂.  tc.el  Copyright 
  ͈ȉ̒ʂł.   
;;
;; T-Code frontend for Nemacs.
;; Author : Yasushi Saito (yasushi@is.s.u-tokyo.ac.jp)
;;          Kaoru Maeda   (kaoru@is.s.u-tokyo.ac.jp)
;;
;; Department of Information Science
;; Faculty of Science, University of Tokyo
;; 7-3-1 Hongo, Bunkyo-ku, 113 Tokyo, Japan
;;
;; Copyright (C) 1989, 1990, 1991 Yasushi Saito and Kaoru Maeda.
;;

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License version 1
;; as published by the Free Software Foundation.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include "comp.h"

#ifdef BUSHU

#define TRUE 1
#define FALSE 0

#define ENTN	7000

struct _dicent {
	unsigned short a, b, c;
};

static struct _dicent *dict;
static int entnum;

#ifdef REVI

#define MAXKC	(95*95)

typedef struct _revp {
	struct _revp *next;
	int index;
} Revp;

static Revp *arev[MAXKC];
static Revp *brev[MAXKC];
static Revp *crev[MAXKC];

#endif /* REVI */

static unsigned int NL;

static int tcode_decompose_char(unsigned int kanji, unsigned int *c1, unsigned int *c2);
static unsigned int tcode_outset_alternative(unsigned int char1);
static unsigned int tcode_lo_sub(unsigned int char1, unsigned int char2);
#ifdef REVI
static void free_rev(void);
unsigned int s2k(unsigned int k);
#endif

/**
 * 񂩂1𓾂B
 * @param s [in/out] B̕ʒuɍXVB
 * @return 擾1Bꍇ0B
 */
moji_t
get_moji(char **s)
{
	moji_t moji;
	char *p = *s;
	if(*p & 0x80) {
		moji = B2S(*p, *(p + 1));
		p++;
	}
	else {
		moji = B2S(0, *p);
	}
	if (*p != '\0')
		p++;
	*s = p;
	return moji;
}

int
read_bushu_dic(char *bushudic)
{
	FILE *f;
#define BUFLEN	1024
	unsigned char buf[BUFLEN];
#ifdef REVI
	int i;
#endif /* REVI */
	int rv = 0;

	NL = B2S(0,'N');
	f=fopen(bushudic,"r");
	if(f==NULL) {
		rv = BE_FOPEN;
		goto ret;
	}
	dict=(struct _dicent*)calloc(ENTN,sizeof(struct _dicent));
	if(!dict) {
		rv = BE_NOMEM;
		goto ret;
	}
	for(entnum=0; fgets(buf, BUFLEN, f) != NULL; entnum++) {
		char *s = buf;
		if(entnum>=ENTN) {
			rv = BE_BIG;
			goto ret;
		}
#ifdef OLD_BUSHU_DIC	/* bushu.dicgꍇ */
		dict[entnum].a = get_moji(&s);
		dict[entnum].b = get_moji(&s);
		dict[entnum].c = get_moji(&s);
#else			/* bushu.revgꍇ */
		dict[entnum].c = get_moji(&s);
		dict[entnum].a = get_moji(&s);
		dict[entnum].b = get_moji(&s);
		/* ̒` (: @A) */
		if(dict[entnum].b == '\r' || dict[entnum].b == '\n') {
			/* Iɂ abc = N@A ƂĎ (`) */
			dict[entnum].b = dict[entnum].c;
			dict[entnum].c = dict[entnum].a;
			dict[entnum].a = NL;
		}
#endif
	}
	fclose(f);
	f = NULL;

#ifdef REVI
	for(i=0; i<MAXKC; i++) {
		arev[i] = brev[i] = crev[i] = NULL;
	}
	for(i=0; i<entnum; i++) {
		Revp *n;
		int k;
		n = (Revp*)calloc(sizeof(Revp),3);
		if(!n) { rv = BE_NOMEM; goto ret; }

		k = s2k(dict[i].a);
		n->index = i; n->next = arev[k]; arev[k] = n;

		n++;
		k = s2k(dict[i].b);
		n->index = i; n->next = brev[k]; brev[k] = n;

		n++;
		k = s2k(dict[i].c);
		n->index = i; n->next = crev[k]; crev[k] = n;
	}

#endif /* REVI */

    ret:
	if (f != NULL)
		fclose(f);
	if (rv != 0) {
		free(dict);
#ifdef REVI
		free_rev();
#endif
	}
	return rv;
}

#ifdef REVI
/**
 * arev, brev, crev̒gfree
 */
static void
free_rev(void)
{
#define FREELIST(list)	for (p = list; p != NULL; p = q) {\
				q = p->next;\
				free(p);\
			}
	int i;

	for (i = 0; i < MAXKC; i++) {
		Revp *p, *q;
		FREELIST(arev[i]);
		FREELIST(brev[i]);
		FREELIST(crev[i]);
	}
}
#endif /* REVI */

/*
 * KANJI2̕ɕB
 *  NULL Ԃ. 
 */
static int
tcode_decompose_char(unsigned int kanji, unsigned int *c1, unsigned int *c2)
{
#ifdef REVI
	Revp *p;
	if((p=crev[s2k(kanji)])!=NULL) {
		*c1 = dict[p->index].a;
		*c2 = dict[p->index].b;
		return TRUE;
	}
#else
	int i;
	for(i=0; i<entnum; i++) {
		if(dict[i].c == kanji) {
			*c1 = dict[i].a;
			*c2 = dict[i].b;
			return TRUE;
		}
	}
#endif
	*c1 = *c2 = 0;
	return FALSE;
}

/*
 * CHAR Ɠȕ EQCHAR ԂBɂȂ CHAR gԂB
 * CHAR łƂ NL + EQCHAR = CHAR ƒ`Ă邱Ƃł
 */
static unsigned int
tcode_outset_alternative(char1)
unsigned int char1;
{
	unsigned int c1, eqchar;

	if(tcode_decompose_char(char1,&c1,&eqchar) && c1==NL) {
		return eqchar;
	}
	return char1;
}

/*
 * TCHAR-1  TCHAR-2 𒼐ڑgݍ킹ĂłO{B
 * tȂꍇɂ NULL Ԃ
 */
static unsigned int
tcode_lo_sub(unsigned int char1, unsigned int char2)
{
	/*
	 * pȂ, Xy[X߂Ă
	 * (if (= (length tchar-1) 1) (setq tchar-1 (concat tchar-1 " ")))
	 * (if (= (length tchar-2) 1) (setq tchar-2 (concat tchar-2 " ")))
	 */
#ifdef REVI
	Revp *p,*q;
	for(p=arev[s2k(char1)]; p!=NULL; p=p->next) {
		for(q=brev[s2k(char2)]; q!=NULL; q=q->next) {
			if(p->index==q->index) {
				return dict[p->index].c;
			}
		}
	}

#else
	int i;
	for(i=0; i<entnum; i++) {
		if((dict[i].a == char1 &&
		    dict[i].b == char2)) {
			/* , char1  char2 Ƃꊷč\
			 * (dict[i].a == char2 && dict[i].b == char1)
			 * Ƃ󋵂͍lȂ
			 */
			return dict[i].c;
		}
	}
#endif
	return 0;
}

/**
 * strŏ2otcode_lookup_outset()ĂԁB
 */
int
tcode_lookup_outset_from_str(char *str)
{
	moji_t moji1, moji2;
	moji1 = get_moji(&str);
	moji2 = get_moji(&str);
	return tcode_lookup_outset(moji1, moji2);
}

#ifndef OKA

/*
 * TCHAR-1  TCHAR-2 gݍ킹ĂłO{B
 * Ȃꍇɂ NULL Ԃ
 */
int
tcode_lookup_outset(
	moji_t	char1,
	moji_t	char2)
{
	unsigned int r;
	unsigned int c11,c12, c21,c22;
	int c;
	unsigned int t;

	/*
	 * ̃[`ł, ̕̕ɖ߂Ȃ悤
	 * ӂȂ΂ȂȂ. 
	 */
#define RETURN(x)	if(x!=char1 && x!=char2) return x
	
	/*
	 * SpŔpɒ镶͔pɂĂ
	 */
	/*
	 * char1 := tcode_2_to_1(tchar-1);
	 * char2 := tcode_2_to_1(tchar-2);
	 */


	for(c=0; c<2; c++) {
		
		/*Ȃł! */
		if(r=tcode_lo_sub(char1,char2)) RETURN(r);
		
		/*
		 * Ȃ
		 * ꂼ̕eqcharɂĂ,
		 */
		char1 = tcode_outset_alternative(char1);
		char2 = tcode_outset_alternative(char2);
		/* eqcharǂōłȂ炨 */
		if(r=tcode_lo_sub(char1,char2)) RETURN(r);
		/*
		 * Ȃ, 
		 * eqcharɕĂ,
		 */
		tcode_decompose_char(char1,&c11,&c12);
		tcode_decompose_char(char2,&c21,&c22);
		
		/* NL͂ */
		if(c11==NL) c11=NULL;
		if(c12==NL) c12=NULL;
		if(c21==NL) c21=NULL;
		if(c22==NL) c22=NULL;
		
		/*
		 * decomposeɃ}b`, Ԃ. 
		 * tc12, tc11, tc22, tc21̂ɕԂ. 
		 */
		if(c11 && c12 && c11==char2) RETURN(c12);
		if(c12 && c11 && c12==char2) RETURN(c11);
		if(c21 && c22 && c21==char1) RETURN(c22);
		if(c22 && c21 && c22==char1) RETURN(c21);
		
		/* ȉ̏ɍł邩ׂ. */
		if(char1 && c22 && (r=tcode_lo_sub(char1,c22))) RETURN(r);
		if(char2 && c12 && (r=tcode_lo_sub(char2,c12))) RETURN(r);
		if(char1 && c21 && (r=tcode_lo_sub(char1,c21))) RETURN(r);
		if(char2 && c11 && (r=tcode_lo_sub(char2,c11))) RETURN(r);
		if(c12 && c22 && (r=tcode_lo_sub(c12,c22))) RETURN(r);
		if(c21 && c12 && (r=tcode_lo_sub(c21,c12))) RETURN(r);
		if(c11 && c22 && (r=tcode_lo_sub(c11,c22))) RETURN(r);
		if(c21 && c11 && (r=tcode_lo_sub(c21,c11))) RETURN(r);
		
		if(c11 && c12 && c11==c21) RETURN(c12);
		if(c11 && c22 && c11==c22) RETURN(c12);
		if(c12 && c21 && c12==c21) RETURN(c11);
		if(c12 && c22 && c12==c22) RETURN(c11);
		
		/* ̏tɂĂĂ݂. */
		t=char1, char1=char2, char2=t;
	}
	return NULL;
}

#else

/*
 * ȉ̃R[h, 
 *  rs (OKA Toshiyuki) oka@nova.co.jp
 * ɂASYp. 
 */

/*
 * TCHAR-1  TCHAR-2 gݍ킹ĂłO{B
 * Ȃꍇɂ NULL Ԃ
 */
int
tcode_lookup_outset(
	moji_t	char1,
	moji_t	char2)
{
	moji_t r;
	moji_t a,b;
	moji_t a1,a2, b1,b2;
	int i;
	moji_t t;

	/*
	 * ̃[`ł, ̕̕ɖ߂Ȃ悤
	 * ӂȂ΂ȂȂ. 
	 */
#define RETURN(x)	if(x!=char1 && x!=char2) return x
	
	/*
	 * SpŔpɒ镶͔pɂĂ
	 */
	/*
	 * char1 := tcode_2_to_1(tchar-1);
	 * char2 := tcode_2_to_1(tchar-2);
	 */

	/*
	 * Ȃ
	 * ꂼ̕eqcharɂĂ,
	 */
	a = tcode_outset_alternative(char1);
	b = tcode_outset_alternative(char2);

	/*
	 * Ȃ, 
	 * eqcharɕĂ,
	 */
	tcode_decompose_char(a,&a1,&a2);
	tcode_decompose_char(b,&b1,&b2);

	/* NL͂ */
	if(a1==NL) a1= 0;
	if(a2==NL) a2= 0;
	if(b1==NL) b1= 0;
	if(b2==NL) b2= 0;
	
	for(i=0; i<2; i++) {
		
		/* ܂Z */
		if((r=tcode_lo_sub(char1,char2))) RETURN(r);
		
		/* eqcharuőZ */
		if((r=tcode_lo_sub(a,b))) RETURN(r);
		
		/* ͈Z */
		if(a2 && a1 && a2==b) RETURN(a1);
		if(a1 && a2 && a1==b) RETURN(a2);

		/* iɂ鑫Z */
		if(a && b1 && (r=tcode_lo_sub(a,b1))) RETURN(r);
		if(a && b2 && (r=tcode_lo_sub(a,b2))) RETURN(r);
		if(a1 && b && (r=tcode_lo_sub(a1,b))) RETURN(r);
		if(a2 && b && (r=tcode_lo_sub(a2,b))) RETURN(r);

		/* iɂ鑫Z */
		if(a1 && b1 && (r=tcode_lo_sub(a1,b1))) RETURN(r);
		if(a1 && b2 && (r=tcode_lo_sub(a1,b2))) RETURN(r);
		if(a2 && b1 && (r=tcode_lo_sub(a2,b1))) RETURN(r);
		if(a2 && b2 && (r=tcode_lo_sub(a2,b2))) RETURN(r);
		
		/* iɂZ */
		if(a2 && b1 && a2==b1) RETURN(a1);
		if(a2 && b2 && a2==b2) RETURN(a1);
		if(a1 && b1 && a1==b1) RETURN(a2);
		if(a1 && b2 && a1==b2) RETURN(a2);
		
		/* ̏tɂĂĂ݂. */
		t=char1, char1=char2, char2=t;
		t=a1, a1=b1, b1=t;
		t=a2, a2=b2, b2=t;
	}
	return 0;
}

#endif

#ifdef REVI

unsigned int
s2k(
	unsigned int k)
{
	unsigned int a,b;
	unsigned int hi,lo;

	a = (k>>8)&0xff;
	b = k&0xff;

	if(a==0 && b=='N') {
		return 0;
	}
	else if(a<0x80) {
		hi = 0;
		lo = (a&0x7f)-0x20;
		return hi*94+lo;
	}
	else {
		hi = (a&0x7f)-0x20;
		lo = (b&0x7f)-0x20;
		return hi*94+lo;
	}
}

#endif /* REVI */
#endif /* BUSHU */
